use Data::Dumper;

package Robot;
use strict;
require Exporter;
use vars qw(@ISA @EXPORT);
@ISA = qw(Exporter);
@EXPORT = qw($sock %robots $board %packages $robot $debug $tot_robots @dead $bot_id);            # symbols to export by default
use IO::Socket::INET;
my $sock;
my %robots;
my $board;
my %packages;
my @homes;
my $robot;
my $bot_id;
my @dead;
my $tot_robots = 1;
my $debug;

sub init_game {
	my ($class, %args) = @_;

	$debug = $args{debug};
  $sock = IO::Socket::INET->new(PeerAddr => $args{addr} || 'icfp1.cse.ogi.edu',
	    PeerPort => $args{port} || '20005',
	    Proto    => 'tcp') || die "couldn't connect";
	print $sock "Player\n";
	my ($x,$y) = split / /, <$sock>;
	print "dimensions $x $y" if $debug;
	$board = Robot::Board->new($x,$y);
	for (1..$y) {
		$board->add_row();
	}
	my @bottom_wall = split //, '#' x ($x+2);
	push @{ $board->{board} }, \@bottom_wall;
	$board->get_homes;
	print Data::Dumper->new([$board],['post_board'])->Dump if $debug;

	my ($uid, $maxCap, $money) = split /\s+/, <$sock>;

	my $raw_robot_data = <$sock>;
	chomp $raw_robot_data;
	$raw_robot_data =~ s/[XY\#]//ig;
	my (@robot_info) = split /\s+/, $raw_robot_data;
	print "raw robot data: $raw_robot_data\n" if $debug;

	while ($#robot_info > 0) {
		my $uid = shift @robot_info;
		my $x   = shift @robot_info;
		my $y   = shift @robot_info;

		$robots{$uid} = Robot->new( uid=>$uid, x=>$x, y=>$y );
	}
	
	$bot_id = $uid;
	$robot  = $robots{$bot_id};
	$robot->{money} = $money;
	$robot->{maxCap}= $maxCap;

	print $\ x 1_000_000;
	$board->draw($board->{board},$robot->{x},$robot->{y}) if $debug;
	print Data::Dumper->new([$robot],['robot'])->Dump if $debug;

	play();
}

sub play {
	my $action;
	while (1) {
		$action = Robot->turn($action);
		my $bid = 2;
		print "$bid $action\n" if $debug;
		print $sock "$bid $action\n";
		my $reply = <$sock>;
		chomp $reply;
		print "reply to move was: $reply\n" if $debug;
		$board->update($reply);
		$board->draw($board->{board},$robot->{x},$robot->{y});
		sleep 1 if $debug;
	}
}

sub new {
	my ($class, %args) = @_;
	warn "new bot uid=$args{uid} x=$args{x} y=$args{y}\n";
	$tot_robots++;
	my $robot = {
		name 	 => chr( $tot_robots+64 ),
		maxCap   => $args{maxCap},
		uid      => $args{uid},
		money    => $args{money},
		x	     	 => $args{x},
		y		 		 => $args{y},
		packages => {},
		score 	 => undef,
	};
	bless $robot, $class;
	return $robot;
}

sub turn {
	my ($class, $action) = @_;

	my (%packages_here) = Robot->get_packages;
	print Data::Dumper->new([%packages_here],['packages_here'])->Dump if $debug > 1;

	# drop off packages that we can
	my @dropoff;
	foreach (keys %{ $robot->{packages} }) {
		my $package = $robot->{packages}->{$_};
		if ($package->{x} == $robot->{x} 
		&&  $package->{y} == $robot->{y}) {
			push @dropoff, $_;
			return 'Drop '. join ' ', @dropoff;
		}
	}
	
	# otherwise pick up if we can
	print Data::Dumper->new([$robot->{packages}],['packages_we_have'])->Dump if $debug;
	my @pickup;
	my $weight = $robot->total_weight;
	foreach my $package (sort { $b->priority <=> $a->priority } values %packages_here) {
		if ($weight + $package->{weight} < $robot->{maxCap}) {
			$weight += $package->{weight};
			print "trying to pickup package $package->{id} with weight $package->{weight} cap: $weight\n" if $debug > 1;
			push @pickup, $package->{id};
		}
	}
	if (scalar @pickup > 0) {
		return 'Pick '. join ' ', @pickup;
	}

	my $myboard = deep_copy($board->{board});
	# or else move
	foreach my $package (sort { $b->priority <=> $a->priority } values %{ $robot->{packages} }) {
		return $robot->go($package->{x},$package->{y},$robot->{x},$robot->{y},$myboard,0);
	}

	# else we don't have any packages to pickup, drop, or move to
	print "closest home is:",$board->home,"\n" if $debug > 1;
	return $robot->go($board->home,$robot->{x},$robot->{y},$myboard,0);
	@dead = ();

	return 'Drop';
}

sub go {
	my ($self, $tx,$ty,$cx,$cy,$myboard,$dir,$depth) = @_;
	print $\ x 1_000_000 if $debug > 1;
	$board->draw($myboard,$cx,$cy) if $debug > 1;
	my $goboard = deep_copy($myboard);
	$goboard->[$cy]->[$cx] = '!';
	foreach (@dead) { $goboard->[$_->[1]]->[$_->[0]] = '*' }
	my %legal = get_legal_moves($tx,$ty,$cx,$cy,$goboard);
	print "legal moves are: ".Data::Dumper->Dumper(%legal)."\n" if $debug > 2;
	return ('Move W') if ($tx == $cx-1 && $ty == $cy);
	return ('Move E') if ($tx == $cx+1 && $ty == $cy); 
	return ('Move S') if ($tx == $cx && $ty == $cy-1);
	return ('Move N') if ($tx == $cx && $ty == $cy+1);
	my $try_bias = 0;
	if ($try_bias) {
		if ($legal{W} && $dir eq 'Move W') {
			return ('Move W');
		}
		if ($legal{E} && $dir eq 'Move E') {
			return ('Move E');
		}
		if ($legal{N} && $dir eq 'Move N') {
			return ('Move N');
		}
		if ($legal{S} && $dir eq 'Move S') {
			return ('Move S');
		}
	}
	if ($legal{E}) {
		return ('Move E') if ($tx > $cx);
	}
	if ($legal{N}) {
		return ('Move N') if ($ty > $cy);
	}
	if ($legal{W}) {
		return ('Move W') if ($tx < $cx);
	}
	if ($legal{S}) {
		return ('Move S') if ($ty < $cy);
	}
	#my @dirs = qw(N S E W);
	#if ($depth > 9) { my $rand = int(rand 4); return 'Move '.$dirs[$rand]; }
	if ($legal{E}) {
		$dir =  $robot->go($tx,$ty,$cx+1,$cy,$goboard,$dir,$depth+1);
	}
	if ($legal{S}) {
		$dir = $robot->go($tx,$ty,$cx,$cy-1,$goboard,$dir,$depth+1);
	}
	if ($legal{W}) {
		$dir = $robot->go($tx,$ty,$cx-1,$cy,$goboard,$dir,$depth+1);
	}
	if ($legal{N}) {
		$dir = $robot->go($tx,$ty,$cx,$cy+1,$goboard,$dir,$depth+1);
	}
	push @dead, [$cx,$cy];
	#return 'Drop';
}

sub go_with_global {
	my ($self, $tx,$ty,$cx,$cy,$myboard,$dir,$depth) = @_;
	print $\ x 1_000_000 if $debug > 1;
	$board->draw($myboard,$cx,$cy) if $debug > 1;
	my $goboard = deep_copy($myboard);
	$goboard->[$cy]->[$cx] = '!';
	foreach (@dead) { $goboard->[$_->[1]]->[$_->[0]] = '*' }
	my %legal = get_legal_moves($tx,$ty,$cx,$cy,$goboard);
	return ('Move W') if ($tx == $cx-1 && $ty == $cy);
	return ('Move E') if ($tx == $cx+1 && $ty == $cy); 
	return ('Move S') if ($tx == $cx && $ty == $cy-1);
	return ('Move N') if ($tx == $cx && $ty == $cy+1);
	if ($legal{W} && $dir eq 'Move W') {
		return ('Move W');
	}
	if ($legal{E} && $dir eq 'Move E') {
		return ('Move E');
	}
	if ($legal{N} && $dir eq 'Move N') {
		return ('Move N');
	}
	if ($legal{S} && $dir eq 'Move S') {
		return ('Move S');
	}
	if ($legal{E}) {
		$dir =  $robot->go_with_global($tx,$ty,$cx+1,$cy,$goboard,$dir,$depth+1);
	}
	if ($legal{S}) {
		$dir = $robot->go_with_global($tx,$ty,$cx,$cy-1,$goboard,$dir,$depth+1);
	}
	if ($legal{W}) {
		$dir = $robot->go_with_global($tx,$ty,$cx-1,$cy,$goboard,$dir,$depth+1);
	}
	if ($legal{N}) {
		$dir = $robot->go_with_global($tx,$ty,$cx,$cy+1,$goboard,$dir,$depth+1);
	}
	push @dead, [$cx,$cy];
	#return 'Drop';
}

sub move {
	my ($self, $direction) = @_;
	if ($direction eq 'N') {
		$self->{y}++;
	}
	elsif ($direction eq 'S') {
		$self->{y}--;
	}
	elsif ($direction eq 'E') {
		$self->{x}++;
	}
	elsif ($direction eq 'W') {
		$self->{x}--;
	}
	print "robot ".$self->{uid}." moved $direction\n" if $debug;
}

sub get_legal_moves {
	my ($tx,$ty,$x,$y,$myboard) = @_;
	#print STDOUT "I'm trying to go to $tx,$ty\n";
	#print STDOUT "I'm at $x,$y\n";
	my %legal;
	if ($myboard->[$y]->[$x+1] =~ /[.\@]/) { $legal{E} = $myboard->[$y]->[$x+1] }
	if ($myboard->[$y]->[$x-1] =~ /[.\@]/) { $legal{W} = $myboard->[$y]->[$x-1] }
	if ($myboard->[$y+1]->[$x] =~ /[.\@]/) { $legal{N} = $myboard->[$y+1]->[$x] }
	if ($myboard->[$y-1]->[$x] =~ /[.\@]/) { $legal{S} = $myboard->[$y-1]->[$x] }
	return %legal;
}

sub total_weight {
	my ($self) = @_;
	my $weight;
	foreach (values %{$self->{packages}}) {
		$weight += $_->{weight};
	}
	return $weight;
}

sub deep_copy {
    my $this = shift;
    if (not ref $this) {
      $this;
    } elsif (ref $this eq "ARRAY") {
      [map deep_copy($_), @$this];
    } elsif (ref $this eq "HASH" || ref $this eq 'Robot::Board') {
      +{map { $_ => deep_copy($this->{$_}) } keys %$this};
    } else { die "what type is $_?" }
}

sub get_packages {
	my $raw_package_data = <$sock>;
	chomp $raw_package_data;
	my (@package_info) = split /\s+/, $raw_package_data;
	print "raw package data: $raw_package_data\n" if $debug > 1;

	my %packages_raw;
	while (scalar @package_info > 0) {
		my $id = shift @package_info;
		$packages_raw{$id} = bless {
			id			=> $id,
			x 			=> shift @package_info,
			y 			=> shift @package_info,
			weight 	=> shift @package_info,
		}, 'Robot::Package';
		$packages{$id} = $packages_raw{$id};
	}
	return %packages_raw;
}

package Robot::Package;

use Robot;

sub priority {
	my ($self) = @_;
	my $d = (abs($robot->{x} - $self->{x}) + 1) * (abs($robot->{y} - $self->{y}) + 1);
	$self->{priority} = $self->{weight} / $d if $d;
	return $self->{priority};
}

package Robot::Board;

use Robot qw($robot);

sub new {
	my ($class, $x, $y) = @_;
	my @top_wall = split //, '#' x ($x+2);
	my @board;
	push @board, \@top_wall;
	my $board = { board => \@board };
	bless $board, $class;
}

sub add_row {
	my ($self) = @_;
	my $input = <$sock>;
	chomp $input;
	$input = '#'.$input.'#';
	my @row = split //, $input;
	push @{ $board->{board} }, \@row;

}

sub get_homes {
	my ($self) = @_;
	my $y;
	foreach (@{ $board->{board} }) {
		$y++;
		my $x;
		foreach (@{ $_ }) { 
			$x++;
			if ($_ eq '@') {
				push @homes, {
					x => $x,
					y	=> $y,
				};
			}
		}
	}
}

sub home {
	my $homed = 100000000;
	my @home;
	foreach (@homes) {
		my $d = (abs($robot->{x} - $_->{x}) + 1) * (abs($robot->{y} - $_->{y}) + 1);
		if ($homed > $d) { @home = ($_->{x}-1,$_->{y}-1); $homed = $d }
	}
	return @home;
}

sub draw {
	my ($self, $myboard, $rx, $ry) = @_;

	my $y;
	foreach (@{ $myboard }) {
		$y++;
		my $x;
		foreach (@{ $_ }) {
			$x++;
			my $botfound;
			my $i;
			foreach (sort keys %robots) {
				$i++;
				if ($robots{$_}->{y}+1 == $y && $robots{$_}->{x}+1 == $x) { print $robots{$_}->{name}; $botfound++ }
			}
			print STDOUT $_ unless $botfound;
		}
		print STDOUT "\n";
	}
}

sub update {
	my ($self, $update) = @_;
	my (@robot_info) = split /#/, $update;
	shift @robot_info;
	my %found_robot;
	foreach (@robot_info) {
		my (@tokens) = split /\s+/, $_;
		my $uid = shift @tokens;
		$found_robot{$uid}++;
		while (scalar @tokens > 0) {
			my $action = shift @tokens;
			print "robot: $_ uid: $uid action $action\n" if $debug;
			if ($action =~ /[NSEW]/) {
				$robots{$uid}->move($action);
			}
			if ($action eq 'P') {
				my $id = shift @tokens;
				$robot->{packages}->{$id} = $packages{$id} if $uid == $robot->{uid};
			}
			if ($action eq 'D') {
				my $id = shift @tokens;
				delete $robot->{packages}->{$id} if $uid == $robot->{uid};
			}
			elsif ($action eq 'X') {
				my $x = shift @tokens;
				shift @tokens;
				my $y = shift @tokens;
				print "new robot at $x,$y\n";
				$robots{$uid} = Robot->new( uid=>$uid, x=>$x, y=>$y ) unless $uid == $robot->{uid};
			}
			#print "remaining tokens: @tokens\n" if $debug ;
		}
	}
	foreach (keys %robots) {
		delete $robots{$_} unless $found_robot{$_};
	}
}


1;

__END__
#death threat
~ A B 	# $board->{board}->{x-1} eq '~' and $board->{board}->{x+1} !~ /'#.~'/
B A ~ 	# $board->{board}->{x+1} eq '~' and $board->{board}->{x-1} !~ /'#.~'/
~
A 	 	# $board->{board}->{y-1} eq '~' and $board->{board}->{x-1} !~ /'#.~'/
B
B
A 	 	# $board->{board}->{y-1} eq '~' and $board->{board}->{x-1} !~ /'#.~'/
~
